from typing import List, Callable, Tuple, Optional, TypeVar, Generic, Any, \
    Generator, Dict, Sequence
import random
import subprocess

T = TypeVar('T')

##############################################################################
# GENERATORS
##############################################################################


Generatable = Any
ValueGenerator = Callable[[], Generatable]


####################
# ATOMS
####################


def integer_gen(min_value: int, max_value: int) -> Callable[[], int]:
    def generator() -> int:
        return random.randrange(min_value, max_value)
    return generator


def pos_integer_gen(max_value: int) -> Callable[[], int]:
    return integer_gen(1, max_value)


def lower_case_let_gen() -> Callable[[], str]:
    def generator() -> str:
        return chr(random.randint(ord('a'), ord('z')))
    return generator


def upper_case_let_gen() -> Callable[[], str]:
    def generator() -> str:
        return chr(random.randint(ord('A'), ord('Z')))
    return generator


def str_gen(generator: Callable[[], str], min_size: int, max_size: int) \
        -> Callable[[], str]:

    char_list_generator = listof_gen(generator, min_size, max_size)

    def generator() -> str:
        return "".join(char_list_generator)
    return generator

####################


def constant_gen(const: T) -> Callable[[], T]:
    def generator() -> T:
        return const
    return generator


def oneof_gen(constants: List[T]) -> Callable[[], T]:
    def generator() -> T:
        return random.choice(constants)
    return generator


def vector_gen(gen: Callable[[], T], size: int) -> Callable[[], List[T]]:
    def generator() -> List[T]:
        res = []
        for i in range(size):
            res.append(gen())
        return res
    return generator


def listof_gen(gen: Callable[[], T], min_size: int,
               max_size: int) -> Callable[[], List[T]]:
    def generator() -> List[T]:
        size = random.randrange(min_size, max_size)
        res = []
        for i in range(size):
            res.append(gen())
        return res
    return generator


def tupleof_gen(generators: List[ValueGenerator]) \
        -> Callable[[], Tuple[Generatable, ...]]:
    def generator() -> Tuple[Generatable, ...]:
        return tuple([x() for x in generators])
    return generator


def eqchoices_gen(choices: List[Callable[[], T]]) \
        -> Callable[[], T]:
    def generator() -> T:
        return random.choice(choices)()
    return generator


def choices_gen(choices: List[Tuple[int, Callable[[], T]]]) \
        -> Callable[[], Tuple[int, T]]:
    size = sum([x[0] for x in choices])

    def generator() -> Tuple[int, T]:
        throw = random.randrange(0, size) - choices[0][0]
        index = 0
        while throw >= 0:
            index += 1
            throw -= choices[index][0]
        return (index, choices[index][1]())

    return generator

##############################################################################
# ARBITRARY
##############################################################################


class Arbitrary(Generic[T]):
    def __init__(self, generator: Callable[[], T]) -> None:
        self.generator = generator

    def get(self) -> T:
        """
        Returns:
            T: one value from the generator provided
        """
        return self.generator()

    def shrink(self, value: T) -> Generator[T, None, None]:
        assert False, "shrink not implemented"
        pass

    def to_str(self, value: T) -> str:
        return str(value)


class ConstantA(Arbitrary[T]):
    """
    Constant arbitrary returns the value provided and shrinks to nothing
    """

    def __init__(self, constant: T):
        super().__init__(constant_gen(constant))

    def shrink(self, value: T) -> Generator[T, None, None]:
        yield from ()


class Unshrinkable(Arbitrary[T]):
    def __init__(self, generator: Arbitrary[T]) -> None:
        super().__init__(generator)

    def shrink(self, value: T) -> Generator[T, None, None]:
        yield from ()


class OneOfA(Arbitrary[T]):
    def __init__(self, constants: List[T]):
        super().__init__(oneof_gen(constants))

    def shrink(self, value: T) -> Generator[T, None, None]:
        yield from ()


class TupleA(Arbitrary[Tuple[Generatable, ...]]):
    """
    Tuple arbitrary returns values combined from the provided arbitraries,
    shrinks to a combination of shrinked values.
    """

    def __init__(self, arbitraries: Sequence[Arbitrary[Generatable]]):
        super().__init__(tupleof_gen([a.generator for a in arbitraries]))
        self.arbitraries = arbitraries

    def combine(self, duos: List[Tuple[Generatable, Arbitrary[Generatable]]],
                buffer: List[Generatable], level: int) \
            -> Generator[Tuple[Generatable, ...], None, None]:

        if(level >= len(duos)):
            yield tuple(buffer)

        for val in duos[level][1].shrink(duos[level][0]):
            buffer[level] = val
            yield from self.combine(duos, buffer, level + 1)

    def shrink(self, value: Tuple[Generatable, ...]) \
            -> Generator[Tuple[Generatable, ...], None, None]:

        for index in range(len(value)):
            tup_list = list(value)
            for new_elem in self.arbitraries[index].shrink(value[index]):
                if new_elem:
                    tup_list[index] = new_elem
                    yield tuple(tup_list)


def permutated_range(start: int, end: int, step: int = 1) -> List[int]:

    res_range = list(range(start, end, step))
    size = len(res_range)

    for i in range(size):
        left = random.randrange(0, size)
        right = random.randrange(0, size)
        res_range[left], res_range[right] = res_range[right], res_range[left]

    return res_range


class ListA(Arbitrary[List[T]]):

    """
    List arbitrary returns list of size in [min_size, max_size) elements
    generated by the provided arbitrary, shrinks to cutdown version of the
    list.
    """

    def __init__(self, arbitrary: Arbitrary[T], min_size: int,
                 max_size: int) -> None:
        super().__init__(listof_gen(arbitrary.generator, min_size, max_size))
        self.arbitrary = arbitrary

    def shrink(self, value: List[T]) -> Generator[List[T], None, None]:
        if not value:
            yield from ()

        for i in permutated_range(0, len(value)):
            yield value[:i] + value[i + 1:]

        # for i in range(len(value)):
        #     for shrunk_value in self.arbitrary.shrink(value[i]):
        #         result = value[:i] + [shrunk_value] + value[i + 1:]
        #         yield result

    def to_str(self, value: List[T]) -> str:
        return str([self.arbitrary.to_str(val) for val in value])


class ChoiceA(Arbitrary[Tuple[int, T]]):
    """
    Choice arbitrary provides tuple of (chosen_index, value) where chosen_index
    is the index of the arbitrary used, value is a value provided to said
    arbitrary, it shrinks to (chosen_index, shrunk_value)
    """

    def __init__(self, choices: Sequence[Tuple[int, Arbitrary[T]]]):
        super().__init__(choices_gen([(v, a.generator) for v, a in choices]))
        self.choices = choices

    def shrink(self, value: Tuple[int, T]) \
            -> Generator[Tuple[int, T], None, None]:
        choice_index, choice_val = value

        for v in self.choices[choice_index][1].shrink(choice_val):
            yield (choice_index, v)

    def to_str(self, value: Tuple[int, T]) -> str:
        return self.choices[value[0]][1].to_str(value[1])


class NumberA(Arbitrary[int]):
    def __init__(self, min_value: int, max_value: int) -> None:
        super().__init__(integer_gen(min_value, max_value))

    def shrink(self, value: int) -> Generator[int, None, None]:
        if(value != 0):
            yield value // 2
            yield abs(value - 1) * (abs(value) // value)


class StringA(ListA[List[str]]):
    def __init__(self, min_value: int, max_value: int) -> None:
        super().__init__(Unshrinkable(eqchoices_gen([lower_case_let_gen(),
                                                     upper_case_let_gen()])),
                         min_value, max_value)

    def to_str(self, value: List[str]) -> str:
        return ''.join(value)

##############################################################################
# TESTING
##############################################################################


class Test(Generic[T]):
    def __init__(self, name: str, arbitrary: Arbitrary[T],
                 value: Optional[T] = None) -> None:
        self.name = name
        self.value = value if value else arbitrary.get()
        self.arbitrary = arbitrary

    def shrink(self) -> Generator['Test[T]', None, None]:
        for shrunk_value in self.arbitrary.shrink(self.value):
            yield Test(self.name, self.arbitrary, shrunk_value)

    def to_str(self) -> str:
        return self.arbitrary.to_str(self.value)


ReducedTest = Tuple[Test[Generatable], int]


class Tester:
    def __init__(self, verbosity: int = 5) -> None:
        self.tests: List[Test[Generatable]] = []
        self.verbosity = verbosity

    def add_test(self, test: Test[Generatable]) -> None:
        self.tests.append(test)

    def run_test(self, test: Test[Generatable]) -> bool:
        assert False, "run_test not implemented"

    def log(self, priority: int, *args, **kwargs) -> None:
        if priority < self.verbosity:
            print(*args, **kwargs)

    def find_smallest(self, test: Test[Generatable]) -> Optional[ReducedTest]:

        if(self.run_test(test)):
            return None

        self.log(1, "Failed")

        shrinks = 0
        smallest = test
        testing = True

        while testing:
            testing = False
            if(shrinks % 10 == 0):
                self.log(2, f"Shrinks {shrinks}")

            self.log(6, smallest.to_str())
            for sub_test in smallest.shrink():
                if not self.run_test(sub_test):
                    testing = True
                    smallest = sub_test
                    shrinks += 1
                    break

        return (smallest, shrinks)

    def run(self) -> List[ReducedTest]:
        failed_tests: List[ReducedTest] = []
        for test in self.tests:
            self.log(1, f"Test start: {test.name}")
            shrunk_version = self.find_smallest(test)
            if(shrunk_version):
                failed_tests.append(shrunk_version)

        return failed_tests

    def save_tests(self) -> None:
        for test in self.tests:
            with open(f"{test.name}.tst", "w") as original_f:
                original_f.write(test.to_str())

    def save_shrunk(self, shrunk_tests: List[ReducedTest]) -> None:
        for (test, shrinks) in shrunk_tests:
            with open(f"{test.name}_red.tst", "w") as shrunk_f:
                shrunk_f.write(test.to_str())


##############################################################################
# REFERENCE IMPLEMENTAION TESTING
##############################################################################

def str_id(x: str) -> str:
    return x


ProcessOutput = Any
ProcessOutputTest = Callable[[ProcessOutput, ProcessOutput], bool]


def str_to_bytes(v: str) -> bytes:
    return bytes(v, 'utf-8')


def bytes_to_str(v: bytes) -> str:
    return v.decode("utf-8")


def stdout_processed_equality(test_preprocessor: Callable[[str], str],
                              exp_preprocessor: Callable[[str], str]) \
        -> Callable[[ProcessOutput, ProcessOutput], bool]:
    def res_tester(test_out: ProcessOutput, exp_out: ProcessOutput) -> bool:
        processed_test_out = test_preprocessor(bytes_to_str(test_out.stdout))
        processed_exp_out = exp_preprocessor(bytes_to_str(exp_out.stdout))
        return processed_test_out == processed_exp_out
    return res_tester


def stdout_equality() -> ProcessOutputTest:
    def res_tester(test_out: ProcessOutput, exp_out: ProcessOutput) -> bool:
        return test_out.stdout == exp_out.stdout
    return res_tester


class RefImplTester(Tester):

    def __init__(self, program_tested: str, program_referenced: str,
                 verbosity: int = 5,
                 stdout_preprocessor: Callable[[str], str] = lambda x: x) \
            -> None:

        super().__init__(verbosity)
        self.program_tested = program_tested
        self.program_referenced = program_referenced
        self.flags: List[str] = []
        self.output_tests: List[ProcessOutputTest] = []

    def set_flags(self, flags: List[str]) -> None:
        self.flags = flags

    def run_test(self, test: Test[Generatable]) -> bool:
        self.log(5, "run_test")
        test_string = test.to_str()
        self.log(6, test_string)

        test_program = subprocess.run(
            [self.program_tested] + self.flags,
            capture_output=True,
            input=str_to_bytes(test_string))

        exp_program = subprocess.run(
            [self.program_referenced] + self.flags,
            capture_output=True,
            input=str_to_bytes(test_string))

        res = True

        for process_test in self.output_tests:
            res = res and process_test(test_program, exp_program)

        self.log(5, f"test result: {('pass' if res else 'fail')}")
        return res

    def add_output_test(self, test: ProcessOutputTest) -> None:
        self.output_tests.append(test)


class Command(TupleA):

    def __init__(self, arbitraries: Sequence[Arbitrary[Generatable]],
                 symbols: Optional[Dict[str, List[int]]] = None) -> None:
        super().__init__(arbitraries)
        self.symbols = symbols if symbols is not None else {}

    def to_str(self, value: Tuple[Generatable, ...]) -> str:
        res = ""
        for i, (v, arb) in enumerate(zip(value, self.arbitraries)):
            res += arb.to_str(v)
            if i < len(value) - 1:
                res += " "
        return res

    def get_symbol(self, symbol: str, value: Tuple[Generatable, ...]) \
            -> List[Generatable]:
        if symbol not in self.symbols:
            return []

        return [value[i] for i in self.symbols[symbol]]


class CommandSequence(ListA[Tuple[int, Generatable]]):
    def __init__(self,
                 commands: List[Tuple[int,
                                      Command]],
                 min_size: int,
                 max_size: int) -> None:
        super().__init__(ChoiceA(commands), min_size, max_size)
        self.commands = commands

    def to_str(self, values: List[Tuple[int, Generatable]]) -> str:
        body = ""
        for i, val in values:
            body += self.arbitrary.to_str((i, val)) + "\n"
        return body

    def get_symbol(self, symbol: str, values: List[Tuple[int, Generatable]]) \
            -> List[Generatable]:
        res = []
        for i, val in values:
            res += self.commands[i][1].get_symbol(symbol, val)
        return res
